part of '../flutterx_live_data.dart';

class MultipleLiveDataBuilder extends StatelessWidget {
  final List<ObservableData> data;
  final LiveWidgetBuilder<List> builder;
  final Widget? placeholder;

  const MultipleLiveDataBuilder({
    super.key,
    required this.data,
    required this.builder,
    this.placeholder,
  });

  static MultipleLiveDataBuilder with2<X1, X2>({
    Key? key,
    required ObservableData<X1> x1,
    required ObservableData<X2> x2,
    required Widget Function(BuildContext context, X1, X2) builder,
    Widget? placeholder,
  }) =>
      MultipleLiveDataBuilder(
          data: [x1, x2],
          builder: (context, values) => builder(context, values[0], values[1]),
          placeholder: placeholder);

  static MultipleLiveDataBuilder with3<X1, X2, X3>({
    Key? key,
    required ObservableData<X1> x1,
    required ObservableData<X2> x2,
    required ObservableData<X3> x3,
    required Widget Function(BuildContext context, X1, X2, X3) builder,
    Widget? placeholder,
  }) =>
      MultipleLiveDataBuilder(
          data: [x1, x2, x3],
          builder: (context, values) => builder(context, values[0], values[1], values[2]),
          placeholder: placeholder);

  static MultipleLiveDataBuilder with4<X1, X2, X3, X4>({
    Key? key,
    required ObservableData<X1> x1,
    required ObservableData<X2> x2,
    required ObservableData<X3> x3,
    required ObservableData<X4> x4,
    required Widget Function(BuildContext context, X1, X2, X3, X4) builder,
    Widget? placeholder,
  }) =>
      MultipleLiveDataBuilder(
          data: [x1, x2, x3, x4],
          builder: (context, values) => builder(context, values[0], values[1], values[2], values[3]),
          placeholder: placeholder);

  static MultipleLiveDataBuilder with5<X1, X2, X3, X4, X5>({
    Key? key,
    required ObservableData<X1> x1,
    required ObservableData<X2> x2,
    required ObservableData<X3> x3,
    required ObservableData<X4> x4,
    required ObservableData<X5> x5,
    required Widget Function(BuildContext context, X1, X2, X3, X4, X5) builder,
    Widget? placeholder,
  }) =>
      MultipleLiveDataBuilder(
          data: [x1, x2, x3, x4, x5],
          builder: (context, values) => builder(context, values[0], values[1], values[2], values[3], values[4]),
          placeholder: placeholder);

  @override
  StatelessElement createElement() => _MultipleLiveDataBuilderElement(this);

  @override
  Widget build(BuildContext context) {
    final values = (context as _MultipleLiveDataBuilderElement)._values;
    return values != null ? builder(context, values) : (placeholder ?? const SizedBox.shrink());
  }

  @override
  void debugFillProperties(DiagnosticPropertiesBuilder properties) => super.debugFillProperties(properties
    ..add(IterableProperty<ObservableData>('data', data))
    ..add(ObjectFlagProperty<LiveWidgetBuilder<List>>.has('builder', builder)));
}

mixin MultipleLiveDataBuilderElementMixin on StatelessElement {
  Set<ObservableData> _sources = {};
  bool _ready = false;

  List<ObservableData> get data;

  @override
  void mount(Element? parent, Object? newSlot) {
    _sources = {...data};
    updateData(_sources);
    for (final source in _sources) source.observeForever(_onChanged, dispatcher: Dispatcher.postFrame);
    super.mount(parent, newSlot);
    _ready = true;
  }

  @override
  void update(StatelessWidget newWidget) {
    _ready = false;
    final delta = {...data};
    updateData(delta);
    for (final source in [...delta]) {
      if (_sources.contains(source))
        delta.remove(source);
      else {
        source.observeForever(_onChanged, dispatcher: Dispatcher.postFrame);
        _sources.add(source);
      }
    }
    for (final source in delta) {
      source.removeObserver(_onChanged);
      _sources.remove(source);
    }
    super.update(newWidget);
    _ready = true;
  }

  @override
  void unmount() {
    for (final source in _sources) source.removeObserver(_onChanged);
    _sources.clear();
    super.unmount();
  }

  @protected
  void updateData(Iterable<ObservableData> sources);

  void _onChanged(_) {
    if (_ready) {
      updateData(_sources);
      markNeedsBuild();
    }
  }
}

class _MultipleLiveDataBuilderElement extends StatelessElement with MultipleLiveDataBuilderElementMixin {
  List? _values;

  @override
  MultipleLiveDataBuilder get widget => super.widget as MultipleLiveDataBuilder;

  @override
  List<ObservableData> get data => widget.data;

  _MultipleLiveDataBuilderElement(MultipleLiveDataBuilder super.widget);

  @override
  void updateData(Iterable<ObservableData> sources) {
    try {
      _values = ObservableData._allValues(sources);
    } catch (_) {
      _values = null; // at least one source is not initialized
    }
  }
}
